"""Generated implementation of view_materialisation_runs."""

# WARNING DO NOT EDIT
# This code was generated from view-materialisation-runs.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..feature_store_run import ExecutionStatistics
from ..jobs import RunError, RunStatus
from ..schedule import ScheduleState
from ..view_materialisation import (
    ViewMaterialisationJobId, ViewMaterialisationJobVersionId,
    ViewMaterialisationRunId
)


@dataclasses.dataclass(frozen=True)
class ViewMaterialisationRunCreationRequest:
    """Request to create a new table caching job run.
    
    Args:
        job (ViewMaterialisationJobId): A data field.
        jobVersion (ViewMaterialisationJobVersionId): A data field.
        scheduleState (typing.Optional[ScheduleState]): A data field.
        status (RunStatus): A data field.
        error (typing.Optional[RunError]): A data field.
        executionStatistics (typing.Optional[ExecutionStatistics]): A data field.
    """
    
    job: ViewMaterialisationJobId
    jobVersion: ViewMaterialisationJobVersionId
    scheduleState: typing.Optional[ScheduleState]
    status: RunStatus
    error: typing.Optional[RunError]
    executionStatistics: typing.Optional[ExecutionStatistics]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ViewMaterialisationRunCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "job": ViewMaterialisationJobId.json_schema(),
                "jobVersion": ViewMaterialisationJobVersionId.json_schema(),
                "scheduleState": {
                    "oneOf": [
                        {"type": "null"},
                        ScheduleState.json_schema(),
                    ]
                },
                "status": RunStatus.json_schema(),
                "error": {
                    "oneOf": [
                        {"type": "null"},
                        RunError.json_schema(),
                    ]
                },
                "executionStatistics": {
                    "oneOf": [
                        {"type": "null"},
                        ExecutionStatistics.json_schema(),
                    ]
                }
            },
            "required": [
                "job",
                "jobVersion",
                "status",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ViewMaterialisationRunCreationRequest:
        """Validate and parse JSON data into an instance of ViewMaterialisationRunCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ViewMaterialisationRunCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ViewMaterialisationRunCreationRequest(
                job=ViewMaterialisationJobId.from_json(data["job"]),
                jobVersion=ViewMaterialisationJobVersionId.from_json(data["jobVersion"]),
                scheduleState=(
                    lambda v: v and ScheduleState.from_json(v)
                )(
                    data.get("scheduleState", None)
                ),
                status=RunStatus.from_json(data["status"]),
                error=(lambda v: v and RunError.from_json(v))(data.get("error", None)),
                executionStatistics=(
                    lambda v: v and ExecutionStatistics.from_json(v)
                )(
                    data.get("executionStatistics", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing ViewMaterialisationRunCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "job": self.job.to_json(),
            "jobVersion": self.jobVersion.to_json(),
            "scheduleState": (lambda v: v and v.to_json())(self.scheduleState),
            "status": self.status.to_json(),
            "error": (lambda v: v and v.to_json())(self.error),
            "executionStatistics": (lambda v: v and v.to_json())(self.executionStatistics)
        }


@dataclasses.dataclass(frozen=True)
class ViewMaterialisationRunUpdateRequest:
    """Request to update a table caching job run.
    
    Args:
        status (RunStatus): A data field.
        error (typing.Optional[RunError]): A data field.
        executionStatistics (typing.Optional[ExecutionStatistics]): A data field.
    """
    
    status: RunStatus
    error: typing.Optional[RunError]
    executionStatistics: typing.Optional[ExecutionStatistics]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ViewMaterialisationRunUpdateRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "status": RunStatus.json_schema(),
                "error": {
                    "oneOf": [
                        {"type": "null"},
                        RunError.json_schema(),
                    ]
                },
                "executionStatistics": {
                    "oneOf": [
                        {"type": "null"},
                        ExecutionStatistics.json_schema(),
                    ]
                }
            },
            "required": [
                "status",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ViewMaterialisationRunUpdateRequest:
        """Validate and parse JSON data into an instance of ViewMaterialisationRunUpdateRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ViewMaterialisationRunUpdateRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ViewMaterialisationRunUpdateRequest(
                status=RunStatus.from_json(data["status"]),
                error=(lambda v: v and RunError.from_json(v))(data.get("error", None)),
                executionStatistics=(
                    lambda v: v and ExecutionStatistics.from_json(v)
                )(
                    data.get("executionStatistics", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing ViewMaterialisationRunUpdateRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "status": self.status.to_json(),
            "error": (lambda v: v and v.to_json())(self.error),
            "executionStatistics": (lambda v: v and v.to_json())(self.executionStatistics)
        }


@dataclasses.dataclass(frozen=True)
class ViewMaterialisationRun:
    """Details of a table caching job run.
    
    Args:
        id (ViewMaterialisationRunId): A data field.
        job (ViewMaterialisationJobId): A data field.
        jobVersion (ViewMaterialisationJobVersionId): A data field.
        scheduleState (typing.Optional[ScheduleState]): A data field.
        status (RunStatus): A data field.
        error (typing.Optional[RunError]): A data field.
        executionStatistics (typing.Optional[ExecutionStatistics]): A data field.
        created (datetime.datetime): A data field.
    """
    
    id: ViewMaterialisationRunId
    job: ViewMaterialisationJobId
    jobVersion: ViewMaterialisationJobVersionId
    scheduleState: typing.Optional[ScheduleState]
    status: RunStatus
    error: typing.Optional[RunError]
    executionStatistics: typing.Optional[ExecutionStatistics]
    created: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ViewMaterialisationRun data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": ViewMaterialisationRunId.json_schema(),
                "job": ViewMaterialisationJobId.json_schema(),
                "jobVersion": ViewMaterialisationJobVersionId.json_schema(),
                "scheduleState": {
                    "oneOf": [
                        {"type": "null"},
                        ScheduleState.json_schema(),
                    ]
                },
                "status": RunStatus.json_schema(),
                "error": {
                    "oneOf": [
                        {"type": "null"},
                        RunError.json_schema(),
                    ]
                },
                "executionStatistics": {
                    "oneOf": [
                        {"type": "null"},
                        ExecutionStatistics.json_schema(),
                    ]
                },
                "created": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "job",
                "jobVersion",
                "status",
                "created",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ViewMaterialisationRun:
        """Validate and parse JSON data into an instance of ViewMaterialisationRun.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ViewMaterialisationRun.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ViewMaterialisationRun(
                id=ViewMaterialisationRunId.from_json(data["id"]),
                job=ViewMaterialisationJobId.from_json(data["job"]),
                jobVersion=ViewMaterialisationJobVersionId.from_json(data["jobVersion"]),
                scheduleState=(
                    lambda v: v and ScheduleState.from_json(v)
                )(
                    data.get("scheduleState", None)
                ),
                status=RunStatus.from_json(data["status"]),
                error=(lambda v: v and RunError.from_json(v))(data.get("error", None)),
                executionStatistics=(
                    lambda v: v and ExecutionStatistics.from_json(v)
                )(
                    data.get("executionStatistics", None)
                ),
                created=isodate.parse_datetime(data["created"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing ViewMaterialisationRun",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "job": self.job.to_json(),
            "jobVersion": self.jobVersion.to_json(),
            "scheduleState": (lambda v: v and v.to_json())(self.scheduleState),
            "status": self.status.to_json(),
            "error": (lambda v: v and v.to_json())(self.error),
            "executionStatistics": (lambda v: v and v.to_json())(self.executionStatistics),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }



"""Generated implementation of ref."""

# WARNING DO NOT EDIT
# This code was generated from ref.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass


@dataclasses.dataclass(frozen=True)
class Ref(abc.ABC):
    """A reference to a branch or commit.
    
    Args:
        ref (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    ref: str
    
    @classmethod
    def json_schema(cls) -> Ref:
        """JSON schema for variant Ref.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Ref:
        """Validate and parse JSON data into an instance of Ref.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Ref.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing Ref", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class CommitRef(Ref):
    """A reference to a specific commit.
    
    Args:
        ref (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "commit"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    ref: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for CommitRef data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "ref": {
                    "type": "string"
                }
            },
            "required": [
                "adt_type",
                "ref",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> CommitRef:
        """Validate and parse JSON data into an instance of CommitRef.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of CommitRef.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return CommitRef(
                ref=str(data["ref"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing CommitRef",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "ref": str(self.ref)
        }


@dataclasses.dataclass(frozen=True)
class BranchRef(Ref):
    """A reference to a branch.
    
    Args:
        ref (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "branch"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    ref: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for BranchRef data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "ref": {
                    "type": "string"
                }
            },
            "required": [
                "adt_type",
                "ref",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> BranchRef:
        """Validate and parse JSON data into an instance of BranchRef.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of BranchRef.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return BranchRef(
                ref=str(data["ref"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing BranchRef",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "ref": str(self.ref)
        }



"""Generated implementation of entity_population."""

# WARNING DO NOT EDIT
# This code was generated from entity-population.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..attribute import Attribute
from ..entity import EntityId
from ..label import Label
from ..table import TableId


@dataclasses.dataclass(frozen=True)
class EntityPopulationId:
    """Unique identifier for an entity population.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EntityPopulationId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EntityPopulationId:
        """Validate and parse JSON data into an instance of EntityPopulationId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EntityPopulationId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EntityPopulationId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing EntityPopulationId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> EntityPopulationId:
        """Parse a JSON string such as a dictionary key."""
        return EntityPopulationId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class EntityPopulationName:
    """Unique name of an entity population.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EntityPopulationName data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EntityPopulationName:
        """Validate and parse JSON data into an instance of EntityPopulationName.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EntityPopulationName.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EntityPopulationName(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing EntityPopulationName", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> EntityPopulationName:
        """Parse a JSON string such as a dictionary key."""
        return EntityPopulationName(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class EntityPopulationVersionId:
    """Unique identifier for versions of an entity population.
    
    Args:
        value (uuid.UUID): A data field.
    """
    
    value: uuid.UUID
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EntityPopulationVersionId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string",
            "format": "uuid"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EntityPopulationVersionId:
        """Validate and parse JSON data into an instance of EntityPopulationVersionId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EntityPopulationVersionId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EntityPopulationVersionId(uuid.UUID(hex=data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing EntityPopulationVersionId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> EntityPopulationVersionId:
        """Parse a JSON string such as a dictionary key."""
        return EntityPopulationVersionId((lambda s: uuid.UUID(hex=s))(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class EntityPopulation:
    """The definition of a set of entities to generate features for and from when.
    
    Args:
        id (EntityPopulationId): A data field.
        name (EntityPopulationName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        entity (EntityId): A data field.
        sources (typing.List[TableId]): A data field.
        expression (str): A data field.
        version (EntityPopulationVersionId): A data field.
    """
    
    id: EntityPopulationId
    name: EntityPopulationName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    entity: EntityId
    sources: typing.List[TableId]
    expression: str
    version: EntityPopulationVersionId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EntityPopulation data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": EntityPopulationId.json_schema(),
                "name": EntityPopulationName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "entity": EntityId.json_schema(),
                "sources": {
                    "type": "array",
                    "item": TableId.json_schema()
                },
                "expression": {
                    "type": "string"
                },
                "version": EntityPopulationVersionId.json_schema()
            },
            "required": [
                "id",
                "name",
                "description",
                "labels",
                "attributes",
                "entity",
                "sources",
                "expression",
                "version",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EntityPopulation:
        """Validate and parse JSON data into an instance of EntityPopulation.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EntityPopulation.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EntityPopulation(
                id=EntityPopulationId.from_json(data["id"]),
                name=EntityPopulationName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                entity=EntityId.from_json(data["entity"]),
                sources=[TableId.from_json(v) for v in data["sources"]],
                expression=str(data["expression"]),
                version=EntityPopulationVersionId.from_json(data["version"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EntityPopulation",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "entity": self.entity.to_json(),
            "sources": [v.to_json() for v in self.sources],
            "expression": str(self.expression),
            "version": self.version.to_json()
        }


@dataclasses.dataclass(frozen=True)
class EntityPopulationCreationRequest:
    """The definition of a set of entities to generate features for and from when.
    
    Args:
        name (EntityPopulationName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        entity (EntityId): A data field.
        sources (typing.List[TableId]): A data field.
        expression (str): A data field.
    """
    
    name: EntityPopulationName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    entity: EntityId
    sources: typing.List[TableId]
    expression: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EntityPopulationCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "name": EntityPopulationName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "entity": EntityId.json_schema(),
                "sources": {
                    "type": "array",
                    "item": TableId.json_schema()
                },
                "expression": {
                    "type": "string"
                }
            },
            "required": [
                "name",
                "description",
                "labels",
                "attributes",
                "entity",
                "sources",
                "expression",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EntityPopulationCreationRequest:
        """Validate and parse JSON data into an instance of EntityPopulationCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EntityPopulationCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EntityPopulationCreationRequest(
                name=EntityPopulationName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                entity=EntityId.from_json(data["entity"]),
                sources=[TableId.from_json(v) for v in data["sources"]],
                expression=str(data["expression"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EntityPopulationCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "entity": self.entity.to_json(),
            "sources": [v.to_json() for v in self.sources],
            "expression": str(self.expression)
        }



"""Generated implementation of projects."""

# WARNING DO NOT EDIT
# This code was generated from projects.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..attribute import Attribute
from ..entity import EntityId
from ..feature_id import FeatureId
from ..feature_set import FeatureSetId
from ..feature_store import FeatureStoreId
from ..label import Label
from ..source import SourceId
from ..table import TableId


@dataclasses.dataclass(frozen=True)
class ProjectId:
    """Unique identifier for a project.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ProjectId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ProjectId:
        """Validate and parse JSON data into an instance of ProjectId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ProjectId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ProjectId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ProjectId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> ProjectId:
        """Parse a JSON string such as a dictionary key."""
        return ProjectId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class ProjectName:
    """Unique name of a project.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ProjectName data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ProjectName:
        """Validate and parse JSON data into an instance of ProjectName.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ProjectName.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ProjectName(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ProjectName", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> ProjectName:
        """Parse a JSON string such as a dictionary key."""
        return ProjectName(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class ProjectVersionId:
    """Unique identifier for versions of a project.
    
    Args:
        value (uuid.UUID): A data field.
    """
    
    value: uuid.UUID
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ProjectVersionId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string",
            "format": "uuid"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ProjectVersionId:
        """Validate and parse JSON data into an instance of ProjectVersionId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ProjectVersionId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ProjectVersionId(uuid.UUID(hex=data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ProjectVersionId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> ProjectVersionId:
        """Parse a JSON string such as a dictionary key."""
        return ProjectVersionId((lambda s: uuid.UUID(hex=s))(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class Project:
    """The definition of a project.
    
    Args:
        id (ProjectId): A data field.
        name (ProjectName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        branches (typing.List[str]): A data field.
        entities (typing.List[EntityId]): A data field.
        features (typing.List[FeatureId]): A data field.
        featureSets (typing.List[FeatureSetId]): A data field.
        featureStores (typing.List[FeatureStoreId]): A data field.
        sources (typing.List[SourceId]): A data field.
        tables (typing.List[TableId]): A data field.
        version (ProjectVersionId): A data field.
    """
    
    id: ProjectId
    name: ProjectName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    branches: typing.List[str]
    entities: typing.List[EntityId]
    features: typing.List[FeatureId]
    featureSets: typing.List[FeatureSetId]
    featureStores: typing.List[FeatureStoreId]
    sources: typing.List[SourceId]
    tables: typing.List[TableId]
    version: ProjectVersionId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Project data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": ProjectId.json_schema(),
                "name": ProjectName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "branches": {
                    "type": "array",
                    "item": {"type": "string"}
                },
                "entities": {
                    "type": "array",
                    "item": EntityId.json_schema()
                },
                "features": {
                    "type": "array",
                    "item": FeatureId.json_schema()
                },
                "featureSets": {
                    "type": "array",
                    "item": FeatureSetId.json_schema()
                },
                "featureStores": {
                    "type": "array",
                    "item": FeatureStoreId.json_schema()
                },
                "sources": {
                    "type": "array",
                    "item": SourceId.json_schema()
                },
                "tables": {
                    "type": "array",
                    "item": TableId.json_schema()
                },
                "version": ProjectVersionId.json_schema()
            },
            "required": [
                "id",
                "name",
                "description",
                "labels",
                "attributes",
                "branches",
                "entities",
                "features",
                "featureSets",
                "featureStores",
                "sources",
                "tables",
                "version",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Project:
        """Validate and parse JSON data into an instance of Project.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Project.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Project(
                id=ProjectId.from_json(data["id"]),
                name=ProjectName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                branches=[str(v) for v in data["branches"]],
                entities=[EntityId.from_json(v) for v in data["entities"]],
                features=[FeatureId.from_json(v) for v in data["features"]],
                featureSets=[FeatureSetId.from_json(v) for v in data["featureSets"]],
                featureStores=[FeatureStoreId.from_json(v) for v in data["featureStores"]],
                sources=[SourceId.from_json(v) for v in data["sources"]],
                tables=[TableId.from_json(v) for v in data["tables"]],
                version=ProjectVersionId.from_json(data["version"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing Project",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "branches": [str(v) for v in self.branches],
            "entities": [v.to_json() for v in self.entities],
            "features": [v.to_json() for v in self.features],
            "featureSets": [v.to_json() for v in self.featureSets],
            "featureStores": [v.to_json() for v in self.featureStores],
            "sources": [v.to_json() for v in self.sources],
            "tables": [v.to_json() for v in self.tables],
            "version": self.version.to_json()
        }


@dataclasses.dataclass(frozen=True)
class ProjectCreationRequest:
    """The definition of a project.
    
    Args:
        name (ProjectName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        branches (typing.List[str]): A data field.
        entities (typing.List[EntityId]): A data field.
        features (typing.List[FeatureId]): A data field.
        featureSets (typing.List[FeatureSetId]): A data field.
        featureStores (typing.List[FeatureStoreId]): A data field.
        sources (typing.List[SourceId]): A data field.
        tables (typing.List[TableId]): A data field.
    """
    
    name: ProjectName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    branches: typing.List[str]
    entities: typing.List[EntityId]
    features: typing.List[FeatureId]
    featureSets: typing.List[FeatureSetId]
    featureStores: typing.List[FeatureStoreId]
    sources: typing.List[SourceId]
    tables: typing.List[TableId]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ProjectCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "name": ProjectName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "branches": {
                    "type": "array",
                    "item": {"type": "string"}
                },
                "entities": {
                    "type": "array",
                    "item": EntityId.json_schema()
                },
                "features": {
                    "type": "array",
                    "item": FeatureId.json_schema()
                },
                "featureSets": {
                    "type": "array",
                    "item": FeatureSetId.json_schema()
                },
                "featureStores": {
                    "type": "array",
                    "item": FeatureStoreId.json_schema()
                },
                "sources": {
                    "type": "array",
                    "item": SourceId.json_schema()
                },
                "tables": {
                    "type": "array",
                    "item": TableId.json_schema()
                }
            },
            "required": [
                "name",
                "description",
                "labels",
                "attributes",
                "branches",
                "entities",
                "features",
                "featureSets",
                "featureStores",
                "sources",
                "tables",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ProjectCreationRequest:
        """Validate and parse JSON data into an instance of ProjectCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ProjectCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ProjectCreationRequest(
                name=ProjectName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                branches=[str(v) for v in data["branches"]],
                entities=[EntityId.from_json(v) for v in data["entities"]],
                features=[FeatureId.from_json(v) for v in data["features"]],
                featureSets=[FeatureSetId.from_json(v) for v in data["featureSets"]],
                featureStores=[FeatureStoreId.from_json(v) for v in data["featureStores"]],
                sources=[SourceId.from_json(v) for v in data["sources"]],
                tables=[TableId.from_json(v) for v in data["tables"]],
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing ProjectCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "branches": [str(v) for v in self.branches],
            "entities": [v.to_json() for v in self.entities],
            "features": [v.to_json() for v in self.features],
            "featureSets": [v.to_json() for v in self.featureSets],
            "featureStores": [v.to_json() for v in self.featureStores],
            "sources": [v.to_json() for v in self.sources],
            "tables": [v.to_json() for v in self.tables]
        }



"""Generated implementation of access_control."""

# WARNING DO NOT EDIT
# This code was generated from access-control.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..user import UserGroupId


@dataclasses.dataclass(frozen=True)
class PrincipalId(abc.ABC):
    """Identifies the subject principal of an access rule.
    
    Args:
        id (UserGroupId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    id: UserGroupId
    
    @classmethod
    def json_schema(cls) -> PrincipalId:
        """JSON schema for variant PrincipalId.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> PrincipalId:
        """Validate and parse JSON data into an instance of PrincipalId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of PrincipalId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing PrincipalId", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class UserGroupIdPrincipalId(PrincipalId):
    """A user group principal.
    
    Args:
        id (UserGroupId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "usergroupid"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    id: UserGroupId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for UserGroupIdPrincipalId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "id": UserGroupId.json_schema()
            },
            "required": [
                "adt_type",
                "id",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserGroupIdPrincipalId:
        """Validate and parse JSON data into an instance of UserGroupIdPrincipalId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserGroupIdPrincipalId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserGroupIdPrincipalId(
                id=UserGroupId.from_json(data["id"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing UserGroupIdPrincipalId",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "id": self.id.to_json()
        }


@dataclasses.dataclass(frozen=True)
class ResourcePattern:
    """Pattern to match path or table within a source or destination.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ResourcePattern data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ResourcePattern:
        """Validate and parse JSON data into an instance of ResourcePattern.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ResourcePattern.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ResourcePattern(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ResourcePattern", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> ResourcePattern:
        """Parse a JSON string such as a dictionary key."""
        return ResourcePattern(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class MaskingRule(abc.ABC):
    """Masking rule.
    
    Args:
        expression (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    expression: str
    
    @classmethod
    def json_schema(cls) -> MaskingRule:
        """JSON schema for variant MaskingRule.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> MaskingRule:
        """Validate and parse JSON data into an instance of MaskingRule.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of MaskingRule.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing MaskingRule", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class Filter(MaskingRule):
    """Row filter.
    
    Args:
        expression (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "filter"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    expression: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Filter data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "expression": {
                    "type": "string"
                }
            },
            "required": [
                "adt_type",
                "expression",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Filter:
        """Validate and parse JSON data into an instance of Filter.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Filter.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Filter(
                expression=str(data["expression"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing Filter",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "expression": str(self.expression)
        }


@dataclasses.dataclass(frozen=True)
class Mask(MaskingRule):
    """Column mask.
    
    Args:
        column (str): A data field.
        expression (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "mask"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    column: str
    expression: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Mask data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "column": {
                    "type": "string"
                },
                "expression": {
                    "type": "string"
                }
            },
            "required": [
                "adt_type",
                "column",
                "expression",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Mask:
        """Validate and parse JSON data into an instance of Mask.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Mask.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Mask(
                column=str(data["column"]),
                expression=str(data["expression"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing Mask",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "column": str(self.column),
            "expression": str(self.expression)
        }


@dataclasses.dataclass(frozen=True)
class AccessRule:
    """Access control rule.
    
    Each access control rule may contain three fields:
    
    1. patterns to match against paths or table names
    2. principals (user groups)
    3. optional masking rules (row filtering or column masking) to be applied for matching resource & principals
    (not used for destinations)
    
    Args:
        resource (ResourcePattern): A data field.
        principals (typing.List[PrincipalId]): A data field.
        maskingRules (typing.Optional[typing.List[MaskingRule]]): A data field.
    """
    
    resource: ResourcePattern
    principals: typing.List[PrincipalId]
    maskingRules: typing.Optional[typing.List[MaskingRule]]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AccessRule data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "resource": ResourcePattern.json_schema(),
                "principals": {
                    "type": "array",
                    "item": PrincipalId.json_schema()
                },
                "maskingRules": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "array", "item": MaskingRule.json_schema()},
                    ]
                }
            },
            "required": [
                "resource",
                "principals",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AccessRule:
        """Validate and parse JSON data into an instance of AccessRule.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AccessRule.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AccessRule(
                resource=ResourcePattern.from_json(data["resource"]),
                principals=[PrincipalId.from_json(v) for v in data["principals"]],
                maskingRules=(
                    lambda v: v and [MaskingRule.from_json(v) for v in v]
                )(
                    data.get("maskingRules", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AccessRule",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "resource": self.resource.to_json(),
            "principals": [v.to_json() for v in self.principals],
            "maskingRules": (lambda v: v and [v.to_json() for v in v])(self.maskingRules)
        }



#
# Copyright 2021 Simple Machines Pty Ltd - All Rights Reserved.
#
# This file is part of Anaml.
#
# Unauthorized copying and/or distribution of this file, via any medium is
# strictly prohibited.
#
"""Domain models for all server interactions. Models support serialisation to/from JSON for REST calls."""



"""Generated implementation of destination_reference."""

# WARNING DO NOT EDIT
# This code was generated from destination-reference.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..destination import DestinationId
from ..kafka_format import KafkaFormat


@dataclasses.dataclass(frozen=True)
class DestinationReference(abc.ABC):
    """A location to store data in a destination data store.
    
    Args:
        destinationId (DestinationId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    destinationId: DestinationId
    
    @classmethod
    def json_schema(cls) -> DestinationReference:
        """JSON schema for variant DestinationReference.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> DestinationReference:
        """Validate and parse JSON data into an instance of DestinationReference.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of DestinationReference.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing DestinationReference", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class FolderDestinationReference(DestinationReference):
    """Store output data in a folder in the given destination.
    
    Args:
        destinationId (DestinationId): A data field.
        folder (str): A data field.
        folderPartitioningEnabled (bool): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "folder"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    destinationId: DestinationId
    folder: str
    folderPartitioningEnabled: bool
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FolderDestinationReference data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "destinationId": DestinationId.json_schema(),
                "folder": {
                    "type": "string"
                },
                "folderPartitioningEnabled": {
                    "type": "boolean"
                }
            },
            "required": [
                "adt_type",
                "destinationId",
                "folder",
                "folderPartitioningEnabled",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FolderDestinationReference:
        """Validate and parse JSON data into an instance of FolderDestinationReference.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FolderDestinationReference.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FolderDestinationReference(
                destinationId=DestinationId.from_json(data["destinationId"]),
                folder=str(data["folder"]),
                folderPartitioningEnabled=bool(data["folderPartitioningEnabled"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing FolderDestinationReference",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "destinationId": self.destinationId.to_json(),
            "folder": str(self.folder),
            "folderPartitioningEnabled": self.folderPartitioningEnabled
        }


@dataclasses.dataclass(frozen=True)
class TableDestinationReference(DestinationReference):
    """Store output data in a table in the given destination.
    
    Args:
        destinationId (DestinationId): A data field.
        tableName (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "table"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    destinationId: DestinationId
    tableName: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableDestinationReference data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "destinationId": DestinationId.json_schema(),
                "tableName": {
                    "type": "string"
                }
            },
            "required": [
                "adt_type",
                "destinationId",
                "tableName",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableDestinationReference:
        """Validate and parse JSON data into an instance of TableDestinationReference.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableDestinationReference.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableDestinationReference(
                destinationId=DestinationId.from_json(data["destinationId"]),
                tableName=str(data["tableName"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TableDestinationReference",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "destinationId": self.destinationId.to_json(),
            "tableName": str(self.tableName)
        }


@dataclasses.dataclass(frozen=True)
class TopicDestinationReference(DestinationReference):
    """Store output data in a topic in the given destination.
    
    Args:
        destinationId (DestinationId): A data field.
        topic (str): A data field.
        format (KafkaFormat): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "topic"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    destinationId: DestinationId
    topic: str
    format: KafkaFormat
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TopicDestinationReference data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "destinationId": DestinationId.json_schema(),
                "topic": {
                    "type": "string"
                },
                "format": KafkaFormat.json_schema()
            },
            "required": [
                "adt_type",
                "destinationId",
                "topic",
                "format",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TopicDestinationReference:
        """Validate and parse JSON data into an instance of TopicDestinationReference.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TopicDestinationReference.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TopicDestinationReference(
                destinationId=DestinationId.from_json(data["destinationId"]),
                topic=str(data["topic"]),
                format=KafkaFormat.from_json(data["format"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TopicDestinationReference",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "destinationId": self.destinationId.to_json(),
            "topic": str(self.topic),
            "format": self.format.to_json()
        }



"""Generated implementation of view_materialisation."""

# WARNING DO NOT EDIT
# This code was generated from view-materialisation.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..attribute import Attribute
from ..cluster import ClusterId
from ..destination_reference import DestinationReference
from ..feature_store import VersionTarget
from ..label import Label
from ..schedule import Schedule
from ..table import TableId


@dataclasses.dataclass(frozen=True)
class ViewMaterialisationJobId:
    """Unique identifier of a table caching job.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ViewMaterialisationJobId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ViewMaterialisationJobId:
        """Validate and parse JSON data into an instance of ViewMaterialisationJobId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ViewMaterialisationJobId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ViewMaterialisationJobId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ViewMaterialisationJobId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> ViewMaterialisationJobId:
        """Parse a JSON string such as a dictionary key."""
        return ViewMaterialisationJobId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class ViewMaterialisationJobName:
    """Unique name for a table caching job.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ViewMaterialisationJobName data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ViewMaterialisationJobName:
        """Validate and parse JSON data into an instance of ViewMaterialisationJobName.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ViewMaterialisationJobName.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ViewMaterialisationJobName(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ViewMaterialisationJobName", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> ViewMaterialisationJobName:
        """Parse a JSON string such as a dictionary key."""
        return ViewMaterialisationJobName(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class ViewMaterialisationJobVersionId:
    """Unique identifier for a specific version of a table caching job.
    
    Args:
        value (uuid.UUID): A data field.
    """
    
    value: uuid.UUID
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ViewMaterialisationJobVersionId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string",
            "format": "uuid"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ViewMaterialisationJobVersionId:
        """Validate and parse JSON data into an instance of ViewMaterialisationJobVersionId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ViewMaterialisationJobVersionId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ViewMaterialisationJobVersionId(uuid.UUID(hex=data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ViewMaterialisationJobVersionId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> ViewMaterialisationJobVersionId:
        """Parse a JSON string such as a dictionary key."""
        return ViewMaterialisationJobVersionId((lambda s: uuid.UUID(hex=s))(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class ViewMaterialisationRunId:
    """Unique identifier of a table caching job run.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ViewMaterialisationRunId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ViewMaterialisationRunId:
        """Validate and parse JSON data into an instance of ViewMaterialisationRunId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ViewMaterialisationRunId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ViewMaterialisationRunId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ViewMaterialisationRunId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return self.value
    
    @classmethod
    def from_json_key(cls, data: str) -> ViewMaterialisationRunId:
        """Parse a JSON string such as a dictionary key."""
        return ViewMaterialisationRunId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class ViewMaterialisation:
    """Details of the table and entity to be cached.
    
    Args:
        table (TableId): A data field.
        destination (DestinationReference): A data field.
    """
    
    table: TableId
    destination: DestinationReference
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ViewMaterialisation data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "table": TableId.json_schema(),
                "destination": DestinationReference.json_schema()
            },
            "required": [
                "table",
                "destination",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ViewMaterialisation:
        """Validate and parse JSON data into an instance of ViewMaterialisation.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ViewMaterialisation.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ViewMaterialisation(
                table=TableId.from_json(data["table"]),
                destination=DestinationReference.from_json(data["destination"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing ViewMaterialisation",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "table": self.table.to_json(),
            "destination": self.destination.to_json()
        }


@dataclasses.dataclass(frozen=True)
class ViewMaterialisationJobCreationRequest(abc.ABC):
    """Request to create a new view materialisation job.
    
    Args:
        attributes (typing.List[Attribute]): A data field.
        cluster (ClusterId): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        name (ViewMaterialisationJobName): A data field.
        versionTarget (typing.Optional[VersionTarget]): A data field.
        views (typing.List[ViewMaterialisation]): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    attributes: typing.List[Attribute]
    cluster: ClusterId
    description: str
    labels: typing.List[Label]
    name: ViewMaterialisationJobName
    versionTarget: typing.Optional[VersionTarget]
    views: typing.List[ViewMaterialisation]
    
    @classmethod
    def json_schema(cls) -> ViewMaterialisationJobCreationRequest:
        """JSON schema for variant ViewMaterialisationJobCreationRequest.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ViewMaterialisationJobCreationRequest:
        """Validate and parse JSON data into an instance of ViewMaterialisationJobCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ViewMaterialisationJobCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ViewMaterialisationJobCreationRequest", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class BatchViewMaterialisationJobCreationRequest(ViewMaterialisationJobCreationRequest):
    """A batch view materialisation job creation request.
    
    Args:
        name (ViewMaterialisationJobName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        views (typing.List[ViewMaterialisation]): A data field.
        schedule (Schedule): A data field.
        cluster (ClusterId): A data field.
        includeMetadata (bool): A data field.
        versionTarget (typing.Optional[VersionTarget]): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "batch"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    name: ViewMaterialisationJobName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    views: typing.List[ViewMaterialisation]
    schedule: Schedule
    cluster: ClusterId
    includeMetadata: bool
    versionTarget: typing.Optional[VersionTarget]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for BatchViewMaterialisationJobCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "name": ViewMaterialisationJobName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "views": {
                    "type": "array",
                    "item": ViewMaterialisation.json_schema()
                },
                "schedule": Schedule.json_schema(),
                "cluster": ClusterId.json_schema(),
                "includeMetadata": {
                    "type": "boolean"
                },
                "versionTarget": {
                    "oneOf": [
                        {"type": "null"},
                        VersionTarget.json_schema(),
                    ]
                }
            },
            "required": [
                "adt_type",
                "name",
                "description",
                "labels",
                "attributes",
                "views",
                "schedule",
                "cluster",
                "includeMetadata",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> BatchViewMaterialisationJobCreationRequest:
        """Validate and parse JSON data into an instance of BatchViewMaterialisationJobCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of BatchViewMaterialisationJobCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return BatchViewMaterialisationJobCreationRequest(
                name=ViewMaterialisationJobName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                views=[ViewMaterialisation.from_json(v) for v in data["views"]],
                schedule=Schedule.from_json(data["schedule"]),
                cluster=ClusterId.from_json(data["cluster"]),
                includeMetadata=bool(data["includeMetadata"]),
                versionTarget=(
                    lambda v: v and VersionTarget.from_json(v)
                )(
                    data.get("versionTarget", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing BatchViewMaterialisationJobCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "views": [v.to_json() for v in self.views],
            "schedule": self.schedule.to_json(),
            "cluster": self.cluster.to_json(),
            "includeMetadata": self.includeMetadata,
            "versionTarget": (lambda v: v and v.to_json())(self.versionTarget)
        }


@dataclasses.dataclass(frozen=True)
class StreamingViewMaterialisationJobCreationRequest(ViewMaterialisationJobCreationRequest):
    """A streaming view materialisation job creation request.
    
    Args:
        name (ViewMaterialisationJobName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        views (typing.List[ViewMaterialisation]): A data field.
        cluster (ClusterId): A data field.
        versionTarget (typing.Optional[VersionTarget]): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "streaming"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    name: ViewMaterialisationJobName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    views: typing.List[ViewMaterialisation]
    cluster: ClusterId
    versionTarget: typing.Optional[VersionTarget]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for StreamingViewMaterialisationJobCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "name": ViewMaterialisationJobName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "views": {
                    "type": "array",
                    "item": ViewMaterialisation.json_schema()
                },
                "cluster": ClusterId.json_schema(),
                "versionTarget": {
                    "oneOf": [
                        {"type": "null"},
                        VersionTarget.json_schema(),
                    ]
                }
            },
            "required": [
                "adt_type",
                "name",
                "description",
                "labels",
                "attributes",
                "views",
                "cluster",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> StreamingViewMaterialisationJobCreationRequest:
        """Validate and parse JSON data into an instance of StreamingViewMaterialisationJobCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of StreamingViewMaterialisationJobCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return StreamingViewMaterialisationJobCreationRequest(
                name=ViewMaterialisationJobName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                views=[ViewMaterialisation.from_json(v) for v in data["views"]],
                cluster=ClusterId.from_json(data["cluster"]),
                versionTarget=(
                    lambda v: v and VersionTarget.from_json(v)
                )(
                    data.get("versionTarget", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing StreamingViewMaterialisationJobCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "views": [v.to_json() for v in self.views],
            "cluster": self.cluster.to_json(),
            "versionTarget": (lambda v: v and v.to_json())(self.versionTarget)
        }


@dataclasses.dataclass(frozen=True)
class ViewMaterialisationJob(abc.ABC):
    """Definition of a view materialisation job.
    
    Args:
        attributes (typing.List[Attribute]): A data field.
        cluster (ClusterId): A data field.
        description (str): A data field.
        id (ViewMaterialisationJobId): A data field.
        labels (typing.List[Label]): A data field.
        name (ViewMaterialisationJobName): A data field.
        version (ViewMaterialisationJobVersionId): A data field.
        versionTarget (typing.Optional[VersionTarget]): A data field.
        views (typing.List[ViewMaterialisation]): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    attributes: typing.List[Attribute]
    cluster: ClusterId
    description: str
    id: ViewMaterialisationJobId
    labels: typing.List[Label]
    name: ViewMaterialisationJobName
    version: ViewMaterialisationJobVersionId
    versionTarget: typing.Optional[VersionTarget]
    views: typing.List[ViewMaterialisation]
    
    @classmethod
    def json_schema(cls) -> ViewMaterialisationJob:
        """JSON schema for variant ViewMaterialisationJob.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ViewMaterialisationJob:
        """Validate and parse JSON data into an instance of ViewMaterialisationJob.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ViewMaterialisationJob.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ViewMaterialisationJob", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class BatchViewMaterialisationJob(ViewMaterialisationJob):
    """A batch view materialisation job.
    
    Args:
        id (ViewMaterialisationJobId): A data field.
        name (ViewMaterialisationJobName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        views (typing.List[ViewMaterialisation]): A data field.
        schedule (Schedule): A data field.
        cluster (ClusterId): A data field.
        includeMetadata (bool): A data field.
        versionTarget (typing.Optional[VersionTarget]): A data field.
        version (ViewMaterialisationJobVersionId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "batch"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    id: ViewMaterialisationJobId
    name: ViewMaterialisationJobName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    views: typing.List[ViewMaterialisation]
    schedule: Schedule
    cluster: ClusterId
    includeMetadata: bool
    versionTarget: typing.Optional[VersionTarget]
    version: ViewMaterialisationJobVersionId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for BatchViewMaterialisationJob data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "id": ViewMaterialisationJobId.json_schema(),
                "name": ViewMaterialisationJobName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "views": {
                    "type": "array",
                    "item": ViewMaterialisation.json_schema()
                },
                "schedule": Schedule.json_schema(),
                "cluster": ClusterId.json_schema(),
                "includeMetadata": {
                    "type": "boolean"
                },
                "versionTarget": {
                    "oneOf": [
                        {"type": "null"},
                        VersionTarget.json_schema(),
                    ]
                },
                "version": ViewMaterialisationJobVersionId.json_schema()
            },
            "required": [
                "adt_type",
                "id",
                "name",
                "description",
                "labels",
                "attributes",
                "views",
                "schedule",
                "cluster",
                "includeMetadata",
                "version",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> BatchViewMaterialisationJob:
        """Validate and parse JSON data into an instance of BatchViewMaterialisationJob.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of BatchViewMaterialisationJob.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return BatchViewMaterialisationJob(
                id=ViewMaterialisationJobId.from_json(data["id"]),
                name=ViewMaterialisationJobName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                views=[ViewMaterialisation.from_json(v) for v in data["views"]],
                schedule=Schedule.from_json(data["schedule"]),
                cluster=ClusterId.from_json(data["cluster"]),
                includeMetadata=bool(data["includeMetadata"]),
                versionTarget=(
                    lambda v: v and VersionTarget.from_json(v)
                )(
                    data.get("versionTarget", None)
                ),
                version=ViewMaterialisationJobVersionId.from_json(data["version"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing BatchViewMaterialisationJob",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "id": self.id.to_json(),
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "views": [v.to_json() for v in self.views],
            "schedule": self.schedule.to_json(),
            "cluster": self.cluster.to_json(),
            "includeMetadata": self.includeMetadata,
            "versionTarget": (lambda v: v and v.to_json())(self.versionTarget),
            "version": self.version.to_json()
        }


@dataclasses.dataclass(frozen=True)
class StreamingViewMaterialisationJob(ViewMaterialisationJob):
    """A streaming view materialisation job.
    
    Args:
        id (ViewMaterialisationJobId): A data field.
        name (ViewMaterialisationJobName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        views (typing.List[ViewMaterialisation]): A data field.
        cluster (ClusterId): A data field.
        versionTarget (typing.Optional[VersionTarget]): A data field.
        version (ViewMaterialisationJobVersionId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "streaming"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    id: ViewMaterialisationJobId
    name: ViewMaterialisationJobName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    views: typing.List[ViewMaterialisation]
    cluster: ClusterId
    versionTarget: typing.Optional[VersionTarget]
    version: ViewMaterialisationJobVersionId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for StreamingViewMaterialisationJob data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "id": ViewMaterialisationJobId.json_schema(),
                "name": ViewMaterialisationJobName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "views": {
                    "type": "array",
                    "item": ViewMaterialisation.json_schema()
                },
                "cluster": ClusterId.json_schema(),
                "versionTarget": {
                    "oneOf": [
                        {"type": "null"},
                        VersionTarget.json_schema(),
                    ]
                },
                "version": ViewMaterialisationJobVersionId.json_schema()
            },
            "required": [
                "adt_type",
                "id",
                "name",
                "description",
                "labels",
                "attributes",
                "views",
                "cluster",
                "version",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> StreamingViewMaterialisationJob:
        """Validate and parse JSON data into an instance of StreamingViewMaterialisationJob.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of StreamingViewMaterialisationJob.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return StreamingViewMaterialisationJob(
                id=ViewMaterialisationJobId.from_json(data["id"]),
                name=ViewMaterialisationJobName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                views=[ViewMaterialisation.from_json(v) for v in data["views"]],
                cluster=ClusterId.from_json(data["cluster"]),
                versionTarget=(
                    lambda v: v and VersionTarget.from_json(v)
                )(
                    data.get("versionTarget", None)
                ),
                version=ViewMaterialisationJobVersionId.from_json(data["version"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing StreamingViewMaterialisationJob",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "id": self.id.to_json(),
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "views": [v.to_json() for v in self.views],
            "cluster": self.cluster.to_json(),
            "versionTarget": (lambda v: v and v.to_json())(self.versionTarget),
            "version": self.version.to_json()
        }



"""Generated implementation of created_updated."""

# WARNING DO NOT EDIT
# This code was generated from created-updated.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..cluster import ClusterId
from ..destination import DestinationId
from ..entity import EntityId
from ..entity_mapping import EntityMappingId
from ..entity_population import EntityPopulationId
from ..event_store import EventStoreId
from ..feature_id import FeatureId
from ..feature_set import FeatureSetId
from ..feature_store import FeatureStoreId
from ..feature_template import TemplateId
from ..jobs import TableMonitoringJobId
from ..source import SourceId
from ..table import TableId
from ..table_caching import TableCachingJobId
from ..view_materialisation import ViewMaterialisationJobId
from ..webhook import WebhookId


@dataclasses.dataclass(frozen=True)
class EntityCreatedUpdated:
    """Entity created and updated timestamps.
    
    Args:
        id (EntityId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: EntityId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EntityCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": EntityId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EntityCreatedUpdated:
        """Validate and parse JSON data into an instance of EntityCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EntityCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EntityCreatedUpdated(
                id=EntityId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EntityCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class EntityMappingCreatedUpdated:
    """Entity Mapping created and updated timestamps.
    
    Args:
        id (EntityMappingId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: EntityMappingId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EntityMappingCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": EntityMappingId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EntityMappingCreatedUpdated:
        """Validate and parse JSON data into an instance of EntityMappingCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EntityMappingCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EntityMappingCreatedUpdated(
                id=EntityMappingId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EntityMappingCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class EntityPopulationCreatedUpdated:
    """Entity Population created and updated timestamps.
    
    Args:
        id (EntityPopulationId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: EntityPopulationId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EntityPopulationCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": EntityPopulationId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EntityPopulationCreatedUpdated:
        """Validate and parse JSON data into an instance of EntityPopulationCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EntityPopulationCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EntityPopulationCreatedUpdated(
                id=EntityPopulationId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EntityPopulationCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class FeatureCreatedUpdated:
    """Feature created and updated timestamps.
    
    Args:
        id (FeatureId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: FeatureId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FeatureCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": FeatureId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureCreatedUpdated:
        """Validate and parse JSON data into an instance of FeatureCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FeatureCreatedUpdated(
                id=FeatureId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing FeatureCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class FeatureSetCreatedUpdated:
    """FeatureSet created and updated timestamps.
    
    Args:
        id (FeatureSetId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: FeatureSetId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FeatureSetCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": FeatureSetId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureSetCreatedUpdated:
        """Validate and parse JSON data into an instance of FeatureSetCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureSetCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FeatureSetCreatedUpdated(
                id=FeatureSetId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing FeatureSetCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class TemplateCreatedUpdated:
    """Template created and updated timestamps.
    
    Args:
        id (TemplateId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: TemplateId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TemplateCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": TemplateId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TemplateCreatedUpdated:
        """Validate and parse JSON data into an instance of TemplateCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TemplateCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TemplateCreatedUpdated(
                id=TemplateId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TemplateCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class TableCreatedUpdated:
    """Table created and updated timestamps.
    
    Args:
        id (TableId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: TableId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": TableId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCreatedUpdated:
        """Validate and parse JSON data into an instance of TableCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCreatedUpdated(
                id=TableId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TableCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class FeatureStoreCreatedUpdated:
    """Feature Store created and updated timestamps.
    
    Args:
        id (FeatureStoreId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: FeatureStoreId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FeatureStoreCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": FeatureStoreId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureStoreCreatedUpdated:
        """Validate and parse JSON data into an instance of FeatureStoreCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureStoreCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FeatureStoreCreatedUpdated(
                id=FeatureStoreId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing FeatureStoreCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class EventStoreCreatedUpdated:
    """Event Store created and updated timestamps.
    
    Args:
        id (EventStoreId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: EventStoreId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EventStoreCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": EventStoreId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EventStoreCreatedUpdated:
        """Validate and parse JSON data into an instance of EventStoreCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EventStoreCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EventStoreCreatedUpdated(
                id=EventStoreId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EventStoreCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class SourceCreatedUpdated:
    """Source created and updated timestamps.
    
    Args:
        id (SourceId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: SourceId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for SourceCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": SourceId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> SourceCreatedUpdated:
        """Validate and parse JSON data into an instance of SourceCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of SourceCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return SourceCreatedUpdated(
                id=SourceId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing SourceCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class DestinationCreatedUpdated:
    """Destination created and updated timestamps.
    
    Args:
        id (DestinationId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: DestinationId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for DestinationCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": DestinationId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> DestinationCreatedUpdated:
        """Validate and parse JSON data into an instance of DestinationCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of DestinationCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return DestinationCreatedUpdated(
                id=DestinationId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing DestinationCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class ClusterCreatedUpdated:
    """Cluster created and updated timestamps.
    
    Args:
        id (ClusterId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: ClusterId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ClusterCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": ClusterId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ClusterCreatedUpdated:
        """Validate and parse JSON data into an instance of ClusterCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ClusterCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ClusterCreatedUpdated(
                id=ClusterId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing ClusterCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class WebhookCreatedUpdated:
    """Webhook created and updated timestamps.
    
    Args:
        id (WebhookId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: WebhookId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for WebhookCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": WebhookId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> WebhookCreatedUpdated:
        """Validate and parse JSON data into an instance of WebhookCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of WebhookCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return WebhookCreatedUpdated(
                id=WebhookId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing WebhookCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class TableMonitoringJobCreatedUpdated:
    """Table Monitoring Job created and updated timestamps.
    
    Args:
        id (TableMonitoringJobId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: TableMonitoringJobId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableMonitoringJobCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": TableMonitoringJobId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableMonitoringJobCreatedUpdated:
        """Validate and parse JSON data into an instance of TableMonitoringJobCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableMonitoringJobCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableMonitoringJobCreatedUpdated(
                id=TableMonitoringJobId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TableMonitoringJobCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class TableCachingJobCreatedUpdated:
    """Table Caching Job created and updated timestamps.
    
    Args:
        id (TableCachingJobId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: TableCachingJobId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableCachingJobCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": TableCachingJobId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableCachingJobCreatedUpdated:
        """Validate and parse JSON data into an instance of TableCachingJobCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableCachingJobCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableCachingJobCreatedUpdated(
                id=TableCachingJobId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing TableCachingJobCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }


@dataclasses.dataclass(frozen=True)
class ViewMaterialisationJobCreatedUpdated:
    """View Materialisation Job created and updated timestamps.
    
    Args:
        id (ViewMaterialisationJobId): A data field.
        created (datetime.datetime): A data field.
        updated (datetime.datetime): A data field.
    """
    
    id: ViewMaterialisationJobId
    created: datetime.datetime
    updated: datetime.datetime
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ViewMaterialisationJobCreatedUpdated data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": ViewMaterialisationJobId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "updated": {
                    "type": "string",
                    "format": "date-time"
                }
            },
            "required": [
                "id",
                "created",
                "updated",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ViewMaterialisationJobCreatedUpdated:
        """Validate and parse JSON data into an instance of ViewMaterialisationJobCreatedUpdated.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ViewMaterialisationJobCreatedUpdated.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ViewMaterialisationJobCreatedUpdated(
                id=ViewMaterialisationJobId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                updated=isodate.parse_datetime(data["updated"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing ViewMaterialisationJobCreatedUpdated",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "updated": self.updated.strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        }



"""Generated implementation of feature_template_creation_request."""

# WARNING DO NOT EDIT
# This code was generated from feature-template-creation-request.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..aggregate import AggregateExpression
from ..attribute import Attribute
from ..entity import EntityId
from ..event_window import EventWindow
from ..feature_id import FeatureId
from ..feature_template import TemplateName
from ..filter_expression import FilterExpression
from ..label import Label
from ..post_aggregate_expression import PostAggregateExpression
from ..select_expression import SelectExpression
from ..table import TableId


@dataclasses.dataclass(frozen=True)
class FeatureTemplateCreationRequest(abc.ABC):
    """Request to create a new feature template.
    
    Args:
        attributes (typing.List[Attribute]): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        name (TemplateName): A data field.
        select (SelectExpression): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    attributes: typing.List[Attribute]
    description: str
    labels: typing.List[Label]
    name: TemplateName
    select: SelectExpression
    
    @classmethod
    def json_schema(cls) -> FeatureTemplateCreationRequest:
        """JSON schema for variant FeatureTemplateCreationRequest.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureTemplateCreationRequest:
        """Validate and parse JSON data into an instance of FeatureTemplateCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureTemplateCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing FeatureTemplateCreationRequest", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class EventFeatureTemplateCreationRequest(FeatureTemplateCreationRequest):
    """Create a new event feature template.
    
    Args:
        name (TemplateName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        table (TableId): A data field.
        window (EventWindow): A data field.
        select (SelectExpression): A data field.
        filter (typing.Optional[FilterExpression]): A data field.
        aggregate (typing.Optional[AggregateExpression]): A data field.
        postAggregateExpr (typing.Optional[PostAggregateExpression]): A data field.
        entityRestrictions (typing.Optional[typing.List[EntityId]]): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "event"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    name: TemplateName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    table: TableId
    window: EventWindow
    select: SelectExpression
    filter: typing.Optional[FilterExpression]
    aggregate: typing.Optional[AggregateExpression]
    postAggregateExpr: typing.Optional[PostAggregateExpression]
    entityRestrictions: typing.Optional[typing.List[EntityId]]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EventFeatureTemplateCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "name": TemplateName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "table": TableId.json_schema(),
                "window": EventWindow.json_schema(),
                "select": SelectExpression.json_schema(),
                "filter": {
                    "oneOf": [
                        {"type": "null"},
                        FilterExpression.json_schema(),
                    ]
                },
                "aggregate": {
                    "oneOf": [
                        {"type": "null"},
                        AggregateExpression.json_schema(),
                    ]
                },
                "postAggregateExpr": {
                    "oneOf": [
                        {"type": "null"},
                        PostAggregateExpression.json_schema(),
                    ]
                },
                "entityRestrictions": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "array", "item": EntityId.json_schema()},
                    ]
                }
            },
            "required": [
                "adt_type",
                "name",
                "description",
                "labels",
                "attributes",
                "table",
                "window",
                "select",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EventFeatureTemplateCreationRequest:
        """Validate and parse JSON data into an instance of EventFeatureTemplateCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EventFeatureTemplateCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EventFeatureTemplateCreationRequest(
                name=TemplateName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                table=TableId.from_json(data["table"]),
                window=EventWindow.from_json(data["window"]),
                select=SelectExpression.from_json(data["select"]),
                filter=(
                    lambda v: v and FilterExpression.from_json(v)
                )(
                    data.get("filter", None)
                ),
                aggregate=(
                    lambda v: v and AggregateExpression.from_json(v)
                )(
                    data.get("aggregate", None)
                ),
                postAggregateExpr=(
                    lambda v: v and PostAggregateExpression.from_json(v)
                )(
                    data.get("postAggregateExpr", None)
                ),
                entityRestrictions=(
                    lambda v: v and [EntityId.from_json(v) for v in v]
                )(
                    data.get("entityRestrictions", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EventFeatureTemplateCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "table": self.table.to_json(),
            "window": self.window.to_json(),
            "select": self.select.to_json(),
            "filter": (lambda v: v and v.to_json())(self.filter),
            "aggregate": (lambda v: v and v.to_json())(self.aggregate),
            "postAggregateExpr": (lambda v: v and v.to_json())(self.postAggregateExpr),
            "entityRestrictions": (lambda v: v and [v.to_json() for v in v])(self.entityRestrictions)
        }


@dataclasses.dataclass(frozen=True)
class RowFeatureTemplateCreationRequest(FeatureTemplateCreationRequest):
    """Create a new row feature template.
    
    Args:
        name (TemplateName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        over (typing.List[FeatureId]): A data field.
        select (SelectExpression): A data field.
        entityId (EntityId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "row"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    name: TemplateName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    over: typing.List[FeatureId]
    select: SelectExpression
    entityId: EntityId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for RowFeatureTemplateCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "name": TemplateName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "over": {
                    "type": "array",
                    "item": FeatureId.json_schema()
                },
                "select": SelectExpression.json_schema(),
                "entityId": EntityId.json_schema()
            },
            "required": [
                "adt_type",
                "name",
                "description",
                "labels",
                "attributes",
                "over",
                "select",
                "entityId",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> RowFeatureTemplateCreationRequest:
        """Validate and parse JSON data into an instance of RowFeatureTemplateCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of RowFeatureTemplateCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return RowFeatureTemplateCreationRequest(
                name=TemplateName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                over=[FeatureId.from_json(v) for v in data["over"]],
                select=SelectExpression.from_json(data["select"]),
                entityId=EntityId.from_json(data["entityId"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing RowFeatureTemplateCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "over": [v.to_json() for v in self.over],
            "select": self.select.to_json(),
            "entityId": self.entityId.to_json()
        }
